#' Plot the counts for a patient.
#'
#' @examples
#' data(patients)
#' patient.plot(countMatrix=patients,patientName="PD18775a")
#'
#' @title patient.plot
#' @param countMatrix count matrix as the one generated by the import.data function.
#' @param patientName name of the patient. This should match a rowname of the countMatrix.
#' @param xlabels boolean value; shall I display x labels?
#' @param freq boolean value; shall I display rates instead of counts?
#' @return A ggplot2 object
#' @export patient.plot
#' @importFrom data.table data.table as.data.table melt :=
#' @import ggplot2
#'
"patient.plot" <- function( countMatrix, patientName, xlabels = TRUE, freq = FALSE ) {
  
  # separate contexts and alterations
  x <- as.data.table(suppressWarnings(melt(as.matrix(countMatrix)[patientName,], value.name = "N")), keep.rownames = "cat")
  x[, Context := paste0(substr(cat, 1,1), ".", substr(cat, 7,7))]
  x[, alt := paste0(substr(cat, 3,3), ">", substr(cat, 5,5))]
  
  # convert to frequency if needed
  if(freq) {
    x[, freq:=N/x[, sum(N)]]
  }
  
  # make the plot
  if(freq) {
    plt <- ggplot(x) + 
      geom_bar(aes(x = Context, y = freq, fill = alt), stat = "identity", position = "identity") + 
      facet_wrap(~alt, nrow=1, scales = "free_x") + 
      theme(axis.text.x = element_text(angle = 90, hjust = 1), panel.background = element_blank(), axis.line = element_line(colour = "black"))+
      theme(legend.position="none")+ggtitle(patientName)+ylab("Frequency of mutations")
  } else {
    plt <- ggplot(x) + 
      geom_bar(aes(x = Context, y = N, fill = alt), stat = "identity", position = "identity") + 
      facet_wrap(~alt, nrow=1, scales = "free_x") + 
      theme(axis.text.x = element_text(angle = 90, hjust = 1), panel.background = element_blank(), axis.line = element_line(colour = "black"))+
      theme(legend.position="none")+ggtitle(patientName)+ylab("Number of mutations")
  }

  if(!xlabels) {
    plt <- plt+theme(axis.text.x=element_blank(), axis.ticks.x=element_blank())
  }

  return(plt)
  
}

#' Plot the discovered signatures.
#'
#' @examples
#' data(nmf_LassoK_example)
#' beta = as.beta(nmf_LassoK_example)
#' signatures.plot(beta=beta)
#'
#' @title signatures.plot
#' @param beta discovered signatures
#' @param useColNames boolean value; shall I use the colnames from beta as names for the signatures?
#' @param mutation_categories if useColNames is FALSE, the trinucleotides categories to be considered can be specified. 
#' An example is provided in the package and can be loaded by data(mutation_categories)
#' @param firstBackground boolean value; shall I display the background signature as the first element?
#' @param xlabels boolean value; shall I display x labels?
#' @return A ggplot2 object
#' @export signatures.plot
#' @importFrom data.table data.table as.data.table melt :=
#' @import ggplot2
#' @import gridExtra
#'
"signatures.plot" <- function( beta, useColNames = TRUE, mutation_categories = NULL, firstBackground = TRUE, xlabels = TRUE ) {
  
  # name the signatures
  if(firstBackground) {
    rownames(beta) <- c("Background", paste0("Signature ", 1:(nrow(beta)-1)))
  }else {
    rownames(beta) <- paste0("Signature ", 1:nrow(beta))
  }
  
  # name the categories if not given
  if(!useColNames) {
    colnames(beta) <- sort(mutation_categories$cat)
  }
  
  # separate context and alteration
  x <- as.data.table(melt(beta, varnames=c("signature", "cat")))
  x[, Context := paste0(substr(cat, 1,1), ".", substr(cat, 7,7))]
  x[, alt := paste0(substr(cat, 3,3), ">", substr(cat, 5,5))]
  
  # make the plot
  glist <- list()
  
  for(i in 1:nrow(beta)) {
    
    plt <- ggplot(x[signature == rownames(beta)[i]]) + 
      geom_bar(aes(x = Context, y = value, fill = alt), stat = "identity", position = "identity") + 
      facet_wrap(~alt, nrow=1, scales = "free_x") + 
      theme(axis.text.x = element_text(angle = 90, hjust = 1), panel.background = element_blank(), axis.line = element_line(colour = "black")) + 
      ggtitle(rownames(beta)[i]) + theme(legend.position="none")+ylab("Frequency of mutations")
    
    if(!xlabels){
      plt <- plt+theme(axis.text.x=element_blank(), axis.ticks.x=element_blank())
    }
    
    glist[[i]] <- plt
  } 
  
  grid.arrange(grobs = glist, ncol=ceiling(nrow(beta)/3))
  
}
